import json

import pytest

from portal.database.models import Course, Project, Review, ReviewItem, Submission, SubmissionState, User, Group
from . import utils


@pytest.fixture
def created_submission(rest_service, client) -> Submission:
    cpp = rest_service.find.course("testcourse1")
    p = cpp.projects[0]

    request_dict = {
        "project_params": {'tags': 'docker'},
        "file_params": {
            "source": {
                "type": "git",
                "url": "https://gitlab.fi.muni.cz/xkompis/test-hello-world.git",
                "branch": "master",
                "checkout": "master"
            }
        }
    }
    request_json = json.dumps(request_dict, cls=utils.DateTimeEncoder)

    response = utils.make_request(client, f"/courses/{cpp.codename}/projects/{p.name}/submissions",
                                  data=request_json,
                                  headers={"content-type": "application/json"}, method='post')
    new_submission = utils.extract_data(response)
    return rest_service.find.submission(new_submission['id'])


def test_list_all_avail(client):
    db_subm = Submission.query.all()
    response = utils.make_request(client, f'/submissions', method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    resp_submissions = utils.extract_data(response)
    assert len(resp_submissions) == len(db_subm)


def test_list_all_avail_for_user(rest_service, client):
    student = rest_service.find.user('student1')
    db_subm = Submission.query.filter(Submission.user == student).all()
    response = utils.make_request(client, f'/submissions?user=student1', method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    resp_submissions = utils.extract_data(response)
    assert len(resp_submissions) == len(db_subm)


def test_list_all_avail_for_course(rest_service, client):
    course = rest_service.find.course('testcourse1')
    db_subm = Submission.query.filter(Submission.course == course).all()
    response = utils.make_request(client, f'/submissions?course=testcourse1', method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    resp_submissions = utils.extract_data(response)
    num_of_submissions = len(resp_submissions)
    assert num_of_submissions == len(db_subm)
    assert num_of_submissions == 4


def test_list_all_avail_for_project(rest_service, client):
    project = rest_service.find.project('testcourse1', 'hw01')
    db_subm = Submission.query.filter(Submission.project == project).all()
    response = utils.make_request(client, f'/submissions?course=testcourse1&project=hw01')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    resp_submissions = utils.extract_data(response)
    assert len(resp_submissions) == len(db_subm)


# TODO: missing some tests (filter by group, role)
def test_list_all_avail_for_group(rest_service, client):
    project = rest_service.find.project('testcourse1', 'hw01')
    group = rest_service.find.group(project.course, 'seminar01')
    db_subm = [sub for sub in Submission.query.filter(Submission.project == project).all() if group in sub.user.groups]
    response = utils.make_request(client, f'/submissions?course=testcourse1&group=seminar01')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    resp_submissions = utils.extract_data(response)
    assert len(resp_submissions) == len(db_subm)


# missing tests for working with zip
def test_read(client):
    submissions = Submission.query.all()
    s = submissions[0]
    response = utils.make_request(client, f'/submissions/{s.id}', method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    submission = utils.extract_data(response)
    utils.assert_submission_in(submissions, submission)


def test_delete(client):
    submissions = Submission.query.all()
    submissions_len = len(submissions)
    s = submissions[0]
    assert s.review
    reviews_len = len(Review.query.all())
    response = utils.make_request(
        client, f'/submissions/{s.id}', method='delete')
    assert response.status_code == 204
    assert response.mimetype == 'application/json'

    submissions_updated = Submission.query.all()
    assert len(submissions_updated) == submissions_len - 1
    assert len(Review.query.all()) == reviews_len - 1


def test_read_state(client):
    submissions = Submission.query.all()
    s = submissions[0]
    response = utils.make_request(
        client, f'/submissions/{s.id}/state', method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    state = json.loads(str(response.get_data().decode("utf-8")))['state']

    assert state == s.state.name


# needs more tests for state transitions
def test_update_state(client):
    submissions = Submission.query.all()
    s = submissions[0]
    request_dict = {
        "state": "CANCELLED"
    }
    request_json = json.dumps(request_dict)

    response = utils.make_request(client, f'/submissions/{s.id}/state', data=request_json,
                                  headers={"content-type": "application/json"}, method='put')
    assert response.status_code == 204
    assert response.mimetype == 'application/json'

    s_updated = Submission.query.all()[0]

    assert s_updated.state == SubmissionState.CANCELLED


# does not pass - needs a Storage mock or config
def test_resubmit(client, created_submission):
    submission = created_submission

    request_dict = {
        "note": "resubmit reason"
    }
    request_json = json.dumps(request_dict)
    response = utils.make_request(client, f'/submissions/{submission.id}/resubmit',
                                  data=request_json,
                                  headers={"content-type": "application/json"}, method='post')

    assert response.status_code == 201
    assert response.mimetype == 'application/json'

    submission = utils.extract_data(response)

    submissions_updated = Submission.query.all()
    utils.assert_submission_in(submissions_updated, submission)


def test_read_review(client):
    submissions = Submission.query.all()
    s = submissions[0]
    assert s.review

    response = utils.make_request(
        client, f'/submissions/{s.id}/review', method='get')
    assert response.status_code == 200
    assert response.mimetype == 'application/json'

    review = utils.extract_data(response)
    utils.assert_review(expected=s.review, actual=review)


def test_create_review(client):
    s = Submission.query.filter_by(state=SubmissionState.IN_PROGRESS).first()
    assert s
    len_reviews = len(Review.query.all())
    len_review_items = len(ReviewItem.query.all())

    request_dict = {
        "review_items": [
            {
                "file": "src/main.c",
                "line": 10,
                "content": "This is wrong!"
            },
            {
                "file": "src.main.c",
                "line": 15,
                "content": "Nice job!"
            }
        ]
    }

    request_json = json.dumps(request_dict)
    response = utils.make_request(client, f'/submissions/{s.id}/review', data=request_json,
                                  headers={"content-type": "application/json"}, method='post')

    assert response.status_code == 201
    assert response.mimetype == 'application/json'

    review = utils.extract_data(response)
    s_updated = Submission.query.filter_by(project_id=s.project_id,
                                           state=SubmissionState.IN_PROGRESS).first()
    assert len(Review.query.all()) == len_reviews + 1
    assert len(ReviewItem.query.all()) == len_review_items + 2
    utils.assert_review(s_updated.review, review)
